// Copyright (c) 2015 .decimal, Inc. All rights reserved.
// Desc:     Worker to perform tasks on thinknode framework

#include "stdafx.h"
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <json/json.h>

#define CURL_STATICLIB
#include <curl/curl.h>

using namespace std;

// API configuration
string basic_user = "ABCDEFGHIJKLMNOPQRSTUVWXYZ123456=";    // Base64 encoded username:password
string api_url = "https://api.thinknode.com/v1.0";          // thinknode url
string app_name = "Dosimetry";                              // app name
string app_version = "1.0.0";                               // app version

// Curl get request call back
static size_t WriteCallback(void *contents, size_t size, size_t nmemb, void *userp)
{
    ((std::string*)userp)->append((char*)contents, size * nmemb);
    return size * nmemb;
}

// Select a specific json tag into string
string get_json_value(string json, string id, int num = 0)
{    
    Json::Value root;
    Json::Value s;
    Json::Reader reader;
    reader.parse(json, root);

    // When there's more than one returned value, use num for index
    if (root.size() == 1 || num == 0) { s = root.get(id, "1"); }
    else { s = root[num].get(id, "1"); }

    return s.asString();
}

// Curllib get http request
string do_curl_get(string auth, string url)
{
    string response;
    long http_code = 0;

    CURL *curl;
    CURLcode res;
    curl = curl_easy_init();

    if (curl) {
        // Header Authentication
        struct curl_slist *headers = NULL;
        headers = curl_slist_append(headers, auth.c_str());
        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
        // Setup response handling
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteCallback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, &response);
        curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
        curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, false); // SSL will need to be turned on

        res = curl_easy_perform(curl);
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &http_code);
        if(res == CURLE_OK && http_code == 200){         
            curl_easy_cleanup(curl);
            return response;
        }
        else {         
            curl_easy_cleanup(curl);
            std::cout << "    HTTP ERROR: " << http_code << std::endl;
            std::exit(2);
        }
    }
    else 
    {
        std::cout << "Unable to curl" << std::endl;
        curl_easy_cleanup(curl);
		return "";
    }
}

// Curllib post http request
string do_curl_post(string auth, string json, string url)
{
    string response;
    long http_code = 0;

    CURL *curl;
    CURLcode res;
    curl = curl_easy_init();

    if (curl) {
        // Header Authentication
        struct curl_slist *headers = NULL;
        headers = curl_slist_append(headers, auth.c_str());
        curl_slist_append(headers, "Content-Type : application/json");

        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
        //curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE_LARGE, strlen(json.c_str()));
        curl_easy_setopt(curl, CURLOPT_POSTFIELDS, json.c_str());
        curl_easy_setopt(curl, CURLOPT_POST, 1L);
        // Setup response handling
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteCallback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, &response);
        curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
        curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, false); // SSL will need to be turned on

        res = curl_easy_perform(curl);
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &http_code);
        if(res == CURLE_OK && http_code == 200){         
            curl_easy_cleanup(curl);
            return response;
        }
        else {         
            curl_easy_cleanup(curl);
            std::cout << "    HTTP ERROR: " << http_code << std::endl;
            std::exit(2);
        }
    }
    else 
    {
        std::cout << "Unable to post curl" << std::endl;
        curl_easy_cleanup(curl);
		return "";
    }
}

// Handles http request to get the user ID from the basic_user
string get_user_token()
{
    std::cout << "Getting User Token..." << std::endl;
    string authentication_string = "Authorization : Basic " + basic_user;
    return get_json_value(do_curl_get(authentication_string, api_url + "/cas/login"), "token");
}

// Handles http request for realm id
string get_realm_id(string token)
{    
    std::cout << "Getting Realm ID..." << std::endl;
    string authentication_string = "Authorization : Bearer " + token;
    return get_json_value(
        do_curl_get(authentication_string, 
        api_url + "/iam/realms"), 
        "id", 
        1); // There's two realms being returned, I want the 2nd
}

// Handles http request for context id
string get_context_id(std::map<string, string> config)
{    
    std::cout << "Getting Context ID..." << std::endl;
    string authentication_string = "Authorization : Bearer " + config["user_token"];
    return get_json_value(
        do_curl_get(authentication_string, 
        api_url + "/iam/realms/" + config["realm_id"] + "/contexts/" + app_name + "/" + app_version), 
        "id");
}

// API Authentication
std::map<string, string> authenticate()
{
    std::map<string, string> config;

    config["user_token"] = get_user_token();
    config["realm_id"] = get_realm_id(config["user_token"]);
    config["context_id"] = get_context_id(config);

    return config;
}

// Grab and post the specified calc request
void post_calc_request()
{
    // Request info
    string path = "C:\\";
    string sjon_request_file = "get_field_rect.json";

    std::map<string, string> iam = authenticate();

    std::ifstream json_file((path + sjon_request_file).c_str());
    string str((std::istreambuf_iterator<char>(json_file)), std::istreambuf_iterator<char>()); 
    
    string authentication_string = "Authorization : Bearer " + iam["user_token"];
    // Get calculation id
    std::cout << "Sending Calculation..." << std::endl;
    string calculation_id = get_json_value(
        do_curl_post(authentication_string, 
        str, 
        api_url + "/calc/?context=" + iam["context_id"]), "id");

    // Get calculation Status - using long polling
    std::cout << "Checking Calculation Status..." << std::endl;
    string calculation_status = get_json_value(
        do_curl_get(authentication_string, 
        api_url + "/calc/" + calculation_id + "/status/?status=completed&progress=1&timeout=5"), 
        "type");
    if (calculation_status.find("failed") != string::npos) 
    { 
        std::cout << "Server Responded: " << calculation_status << std::endl; 
        return; 
    }

    // Get calculation Result
    std::cout << "Fetching Calculation Result..." << std::endl;
    string calculation_result = do_curl_get(
        authentication_string, 
        api_url + "/calc/" + calculation_id + "/result/?context=" + iam["context_id"]);

    std::cout << "Calculation Result: " << calculation_result << std::endl;
}

// Grab and post sepcified object to the ISS
void post_immutable_object()
{
    // Immutable info
    string path = "C:\\";
    string sjon_iss_file = "aperture.json";
    string obj_name = "aperture";

    std::map<string, string> iam = authenticate();

    // Read local immutable json file
    std::ifstream json_file((path + sjon_iss_file).c_str());
    string str((std::istreambuf_iterator<char>(json_file)), std::istreambuf_iterator<char>());    
    
    // Post object
    std::cout << "Posting Object to ISS..." << std::endl;
    string authentication_string = "Authorization : Bearer " + iam["user_token"];
    string res = do_curl_post(
        authentication_string, 
        str, 
        api_url + "/iss/named/" + app_name + "/" + obj_name + "/?context=" + iam["context_id"]);
    std::cout << "Immuntable ID: " << res << std::endl;   
}

int main(void)
{
    // Which task to do:
    //post_immutable_object();
    //post_calc_request();

	return 0;
}
